require("simulation")
gridrenderoffset = {x = 50, y = 70}
parrendersize = 5


function love.load()
    simulation.init()    
    love.window.setMode( 1280, 720,{resizable=true} )
    love.window.setTitle("Sand simulation")
end



function love.draw()
    love.graphics.setColor(1,1,1,1)
    love.graphics.print("Size: [" .. simulation.size.x .. "," .. simulation.size.y .. "]", 0, 0)
    love.graphics.print("Inputmode: " .. input.mode, 0, 20)
    love.graphics.print("SAND: " .. simulation.count_type(simulation.SAND) .. ", WATER: " .. simulation.count_type(simulation.WATER) .. 
        ", CONCRETE: " .. simulation.count_type(simulation.CONCRETE) .. ", PLANTS: " .. simulation.count_type(simulation.PLANT), 
        100,0)

    rendersimulation()
end


function love.update(dt)
    input.update()
    simulation.update(dt)
end

function rendersimulation( )
    local grid = simulation.grid;
    love.graphics.setColor(1,1,1,1)
    love.graphics.line(gridrenderoffset.x, gridrenderoffset.y, gridrenderoffset.x + simulation.size.x * parrendersize,gridrenderoffset.y)
    love.graphics.line(gridrenderoffset.x, gridrenderoffset.y, gridrenderoffset.x, gridrenderoffset.y + simulation.size.y * parrendersize)
    love.graphics.line(gridrenderoffset.x, gridrenderoffset.y + simulation.size.y * parrendersize, 
        gridrenderoffset.x + simulation.size.x * parrendersize, gridrenderoffset.y + simulation.size.y * parrendersize)
    love.graphics.line(gridrenderoffset.x + simulation.size.x * parrendersize + parrendersize, gridrenderoffset.y, 
        gridrenderoffset.x + simulation.size.x * parrendersize + parrendersize, gridrenderoffset.y + simulation.size.y * parrendersize)
    for x=1,simulation.size.x do
        for y=simulation.size.y,1,-1 do
            local currentCell = simulation.grid[x][y]
            local paricletype = simulation.particletypes[currentCell]
            love.graphics.setColor(paricletype.color)
            love.graphics.rectangle("fill",gridrenderoffset.x+(x* parrendersize), gridrenderoffset.y + (simulation.size.y * parrendersize) - (y * parrendersize),parrendersize,parrendersize)
        end
    end
end

function love.mousepressed(x,y,button)
    input.onmousepressed(x,y,button)
end

function love.mousemoved(x,y)
    input.onmousemove(x,y)
end

function love.mousereleased()
    input.mousedown = false
end

function love.wheelmoved( x, y )
    if y > 0 and parrendersize > 1 then
        parrendersize = parrendersize -1
    elseif  y < 0 and parrendersize <= 10 then
    parrendersize = parrendersize +1
    end
end


function love.keypressed( key )
    if(key == "1") then
        input.mode = "SAND"
    elseif (key == "2") then
        input.mode = "WATER"
    elseif (key == "3") then
        input.mode = "CONCRETE"
    elseif (key == "4") then
        input.mode = "PLANT"
    elseif (key == "5") then
        input.mode = "ERASER"
    elseif (key == "6") then
        input.mode = "FIRE"
    end
end

input = {}
input.mousedown = false
input.mode = "SAND"
input.mousecoords = {x = 0, y = 0}
function input.onmousepressed(x,y,button)
 if(mouseinsimulationwindow(x,y)) then
        local simucoords = input.mouse_to_simu_cords(x,y)

        input.spawn(simucoords.x,simucoords.y)
    end
    input.mousecoords.x = x
    input.mousecoords.y = y
    input.mousedown = true

end

function input.onmousemove(x,y)
    if(input.mousedown == false or not mouseinsimulationwindow(x,y)) then
        return
    end
    local simucoords = input.mouse_to_simu_cords(x,y)
    input.mousecoords.x = x
    input.mousecoords.y = y
end

function input.update( )
    if input.mousedown then
        local simucoords = input.mouse_to_simu_cords(input.mousecoords.x,input.mousecoords.y)
        input.spawn(simucoords.x,simucoords.y)
    end
end

function input.spawn( x,y )
    if(input.mode == "SAND") then
         simulation.spawnsand(x,y)
    elseif input.mode == "WATER" then
         simulation.spawnwater(x,y)
    elseif input.mode == "CONCRETE" then
         simulation.spawnconcrete(x,y)
    elseif input.mode == "PLANT" then
        simulation.spawnplant(x,y)
    elseif input.mode == "ERASER" then
        simulation.erase(x,y) 
    elseif input.mode == "FIRE" then
        simulation.spawnfire(x,y)
    end
end

function input.mouse_to_simu_cords(mouse_x,mouse_y)
    return {x = math.floor((mouse_x - gridrenderoffset.x) /parrendersize) , 
        y = math.floor(simulation.size.y - ((mouse_y - gridrenderoffset.y) / parrendersize))}
end

function mouseinsimulationwindow( x,y )
    return x >= gridrenderoffset.x and y >= gridrenderoffset.y and x <= (gridrenderoffset.x+simulation.size.x * parrendersize) and y <= gridrenderoffset.y+simulation.size.y * parrendersize
end


