simulation = {}
simulation.size = {x = 200, y  = 130}
simulation.grid = {}
simulation._shadowgrid = {}
simulation.speed = 33 -- 1 = once per second, 2 = twice per second, 33 = 33 times per second
simulation.gravity = {x = 0, y = -1}
simulation.plantgrowtime = 20
simulation.defaultfuel = 45
simulation.defaultfirehp = 10
simulation.smokeammount = 5 -- 5 = every fifth tick

simulation.currenttick = 0

simulation.NOTHING = 1
simulation.SAND = 2
simulation.WATER = 3
simulation.CONCRETE = 4
simulation.PLANT = 5
simulation.FIRE = 6
simulation.SMOKE = 7

simulation.particletypes = {{id=1, name = "Nothing",color = {0,0,0,1}},
{id =2,name="Sand",color = {194 / 255, 178 / 255, 128 / 255, 1}},
{id=3,name="water",color = {0 / 255, 94 / 255, 184 / 255, 1}},
{id=4,name="concrete",color = {125 / 255, 125 / 255, 125 / 255, 1}},
{id=5,name="Plant",color = {90 / 255, 171 / 255, 86 / 255, 1}},
{id=6,name="Fire",color = {255 / 255, 122 / 255, 0 / 255, 1}},
{id=7,name="Smoke",color = {170 / 255, 170 / 255, 170 / 255, 1}},}

simulation.water_indices_to_check = {}

simulation.additionalinfo = {}

function simulation.init()
	for x = 1,simulation.size.x do
		simulation.grid[x] = {}
		simulation._shadowgrid[x] = {}
		simulation.additionalinfo[x] = {}
		for y=1,simulation.size.y do
			simulation.grid[x][y] = simulation.NOTHING
			simulation._shadowgrid[x][y] = simulation.grid[x][y]
			simulation.additionalinfo[x][y] = {}
		end
	end
	

	simulation._speed = 1 / simulation.speed
	simulation.testinit()

	simulation.updateshadowgrid()
	simulation.init_water_indices()
end

function simulation.testinit()
	for y = 30,80 do
		simulation.grid[10][y] = simulation.SAND
	end

	for x = 1,simulation.size.x do
		simulation.grid[x][1] = simulation.CONCRETE
	end
end



function simulation.init_water_indices()
	-- precalculating the order in which the water cells should check for free space.
	-- we don't have to recalculate that every frame for every posible position
	for i=1,simulation.size.x do
		simulation.water_indices_to_check[i] = {}
		for j=1,simulation.size.x do
			table.insert(simulation.water_indices_to_check[i],j)
		end
		table.sort(simulation.water_indices_to_check[i], function (n1,n2)
			return math.abs(i - n1) < math.abs(i - n2)
		end)
	end
	
end


function simulation.update(dt)
	simulation.updateshadowgrid()

	simulation.currenttick = simulation.currenttick + dt
	if simulation.currenttick>=simulation._speed then
		simulation._actualupdate()
		simulation.currenttick = 0
	end
end

function simulation._actualupdate()
	-- actually do something 
		for x = 1,simulation.size.x do
			for y=1,simulation.size.y do
				if not simulation.isdirty(x,y) then
					if simulation.grid[x][y] == simulation.SAND then
						simulation.updatesand(x,y)
					elseif simulation.grid[x][y] == simulation.WATER then
						simulation.updatewater(x,y)
					elseif simulation.grid[x][y] == simulation.CONCRETE then
						simulation.updateconcrete(x,y)
					elseif simulation.grid[x][y] == simulation.PLANT then
						simulation.updateplants(x,y)
					elseif simulation.grid[x][y] == simulation.FIRE then
						simulation.updatefire(x,y)
					elseif simulation.grid[x][y] == simulation.SMOKE then
						simulation.updatesmoke(x,y)
					end
				end
			end
		end
	simulation.applyupdates()
end

function simulation.isdirty(x,y)
	return simulation.grid[x][y] ~= simulation._shadowgrid[x][y]
end

function simulation.updateshadowgrid()
	for x = 1,simulation.size.x do
		for y=1,simulation.size.y do
			simulation._shadowgrid[x][y] = simulation.grid[x][y]
		end
	end
end

function simulation.applyupdates()
	for x = 1,simulation.size.x do
		for y=1,simulation.size.y do
			simulation.grid[x][y] = simulation._shadowgrid[x][y]
		end
	end
end

function simulation.updatesand(x,y)
	local gravity = simulation.gravity
	if simulation.free_of_solids(x+gravity.x,y+gravity.y) then
		-- direkt unter mir ist frei
		simulation.transferto(x,y,x+gravity.x,y+gravity.y)
	elseif simulation.free_of_solids(x+gravity.x - 1,y + gravity.y) then
		-- links unter mir ist frei
		simulation.transferto(x,y,x+gravity.x-1,y + gravity.y)
	elseif simulation.free_of_solids(x + gravity.x + 1,y + gravity.y) then
		-- rechts unter mir ist frei
		simulation.transferto(x,y,x+gravity.x+1,y + gravity.y)
	end
end
function simulation.updatewater(x,y)
	local gravity = simulation.gravity
	if simulation.free(x+gravity.x,y+gravity.y) then
		-- direkt unter mir ist frei
		simulation.transferto(x,y,x+gravity.x,y+gravity.y)
		return
	end

	if simulation.istype(x,y+gravity.y,simulation.FIRE) then
		simulation.setto(x,y+gravity.y,simulation.NOTHING)
		simulation.transferto(x,y,x+gravity.x,y+gravity.y)
		return
	end
	
	local wall_right=false
	local wall_left=false
	for i,v in ipairs(simulation.water_indices_to_check[x]) do
		if (v<x and not wall_left) or (v > x and not wall_right) then
			local direction = 0
			if v < x then direction = -1 else direction = 1 end
			if simulation.free(v,y+gravity.y) and simulation.free(x+direction,y) then
				if simulation.istype(x,y-1,simulation.WATER) then
					simulation.transferto(x,y,v,y-1)
				else
					simulation.transferto(x,y,x + direction,y)
				end
				return
			elseif not simulation.free_of_solids(v,y) then
				if v<x then
					wall_left = true
				elseif v > x then
					wall_right = true
				end
			end

			if wall_left and wall_right then
				return
			end
		end
	end

end

function simulation.updateconcrete(x,y)
	-- currently we do nothing becaus concrete is magic and immovable
	return
end

function simulation.updateplants(x,y)
	local gravity = simulation.gravity
	-- nur direkt runter fallen
	if simulation.free_of_solids(x+gravity.x,y+gravity.y) and not simulation.istype(x+gravity.x,y+gravity.y,simulation.FIRE) then
		-- direkt unter mir ist frei
		simulation.transferto(x,y,x+gravity.x,y+gravity.y)
		return
	end

	if simulation.additionalinfo[x][y].growthtimer == nil then
		simulation.additionalinfo[x][y].growthtimer = simulation.plantgrowtime * 2
	end

	if(simulation.additionalinfo[x][y].growthtimer <= 0) then
		if(simulation.viable_for_plant_growth(x-1,y)) then
			simulation.setto(x-1,y,simulation.PLANT)
		elseif(simulation.viable_for_plant_growth(x,y+1)) then
			simulation.setto(x,y+1,simulation.PLANT)
		elseif(simulation.viable_for_plant_growth(x+1,y)) then
			simulation.setto(x+1,y,simulation.PLANT)
		elseif(simulation.viable_for_plant_growth(x+1,y-1)) then
			simulation.setto(x+1,y-1,simulation.PLANT)
		elseif(simulation.viable_for_plant_growth(x-1,y-1)) then
			simulation.setto(x-1,y-1,simulation.PLANT)
		end

		simulation.additionalinfo[x][y].growthtimer  = simulation.plantgrowtime
	else
		simulation.additionalinfo[x][y].growthtimer = simulation.additionalinfo[x][y].growthtimer -1
	end

end

function simulation.viable_for_plant_growth(x,y)
	return simulation.free(x,y) and simulation.solid(x,y -1) and not (simulation.istype(x,y-1,simulation.PLANT) and simulation.istype(x,y-2,simulation.PLANT) and simulation.istype(x,y-3,simulation.PLANT)) 
end

function simulation.updatefire(x,y)
	if(simulation.additionalinfo_getfuel(x,y) <= 0) then
		simulation.setto(x,y,simulation.NOTHING)
		return
	end

	if simulation.free(x,y+1) and simulation.additionalinfo_getsmoketimer(x,y) % simulation.smokeammount == 0 then
		if math.random() < 0.5 then
			simulation.setto(x,y+1,simulation.SMOKE) 
		end
	end
	simulation.additionalinfo_updatesmoketimer(x,y,1)

	
	simulation.tryignite(x-1,y)
	simulation.tryignite(x-1,y-1)
	simulation.tryignite(x,y-1)
	simulation.tryignite(x+1,y-1)
	simulation.tryignite(x+1,y)
	simulation.tryignite(x+1,y+1)
	simulation.tryignite(x,y+1)
	simulation.tryignite(x-1,y+1)

	simulation.additionalinfo[x][y].fuel = simulation.additionalinfo_getfuel(x,y) -1

end

function simulation.tryignite(x,y)
	if simulation.flammable(x,y) then
		if simulation.additionalinfo_getfirehp(x,y) > 0 then
			simulation.additionalinfo[x][y].firehp = simulation.additionalinfo_getfirehp(x,y) -1
		else
			--Turn to fire
			simulation.setto(x,y,simulation.FIRE)
			simulation.additionalinfo[x][y].fuel = simulation.defaultfuel
		end
	end
end


function simulation.additionalinfo_getfuel(x,y) 
	if simulation.additionalinfo[x][y].fuel == nil and simulation.istype(x,y,simulation.FIRE) then
			simulation.additionalinfo[x][y].fuel = simulation.defaultfuel	
	end
	
	return simulation.additionalinfo[x][y].fuel
	
end

function simulation.additionalinfo_getsmoketimer(x,y) 
	if simulation.additionalinfo[x][y].smoketimer == nil and simulation.istype(x,y,simulation.FIRE) then
			simulation.additionalinfo[x][y].smoketimer = 0	
	end
	
	return simulation.additionalinfo[x][y].smoketimer
	
end

function simulation.additionalinfo_updatesmoketimer(x,y,ammount)
	if simulation.additionalinfo[x][y].smoketimer == nil then
		simulation.additionalinfo[x][y].smoketimer = 0
	else
		simulation.additionalinfo[x][y].smoketimer = simulation.additionalinfo[x][y].smoketimer + ammount
	end

end

function simulation.additionalinfo_getfirehp(x,y) 
	if simulation.additionalinfo[x][y].firehp == nil and simulation.flammable(x,y) then
		simulation.additionalinfo[x][y].firehp = simulation.defaultfirehp
	end
	
	return simulation.additionalinfo[x][y].firehp
	
end

function simulation.updatesmoke(x,y)
	if(simulation.free(x,y+1)) then
		simulation.transferto(x,y,x-simulation.gravity.x,y-simulation.gravity.y)
		return
	end

	local wall_right=false
	local wall_left=false
	for i,v in ipairs(simulation.water_indices_to_check[x]) do
		if (v<x and not wall_left) or (v > x and not wall_right) then
			local direction = 0
			if v < x then direction = -1 else direction = 1 end
			if simulation.free(v,y-simulation.gravity.y) and simulation.free(x+direction,y) then
				simulation.transferto(x,y,x + direction,y)
				return
			elseif not simulation.free_of_solids(v,y) then
				if v<x then
					wall_left = true
				elseif v > x then
					wall_right = true
				end
			end

			if wall_left and wall_right then
				simulation.setto(x,y,simulation.NOTHING)
				return
			end
		end
	end

	simulation.setto(x,y,simulation.NOTHING)
	
end


function simulation.inbounds(x,y)
	return x > 0 and x < simulation.size.x and y > 0 and y < simulation.size.y
end


function simulation.transferto(oldX,oldY,newX,newY)
	local typebefore = simulation._shadowgrid[newX][newY]
	simulation._shadowgrid[newX][newY] = simulation.grid[oldX][oldY]
	simulation._shadowgrid[oldX][oldY] = typebefore
end

function simulation.setto(newX,newY,type)
	simulation._shadowgrid[newX][newY] = type
end

function simulation.istype(x,y,type)
	if not simulation.inbounds(x,y) then
		return false
	end
	return simulation._shadowgrid[x][y] == type
end

function simulation.free(x,y)
	if not simulation.inbounds(x,y) then
		return false
	end
	return simulation._shadowgrid[x][y] == simulation.NOTHING
end

function simulation.free_of_solids(x,y)
	if not simulation.inbounds(x,y) then
		return false
	end
	return not simulation._unsavesolid(x,y)
end

function simulation.solid(x,y)
	if not simulation.inbounds(x,y) then
		return false
	end
	return simulation._unsavesolid(x,y)
end

function simulation._unsavesolid(x, y)
	return (simulation._shadowgrid[x][y] == simulation.SAND) or (simulation._shadowgrid[x][y] == simulation.CONCRETE) or (simulation._shadowgrid[x][y] == simulation.PLANT)
end

function simulation.flammable(x,y)
	if not simulation.inbounds(x,y) then
		return false
	end
	return simulation._shadowgrid[x][y] == simulation.PLANT
end


function simulation.spawnsand(x,y) 
	if not simulation.inbounds(x,y) then
		return
	end
	simulation.grid[x][y] = simulation.SAND
end


function simulation.spawnwater(x,y) 
	if not simulation.inbounds(x,y) then
		return
	end
	simulation.grid[x][y] = simulation.WATER
end

function simulation.spawnconcrete(x,y) 
	if not simulation.inbounds(x,y) then
		return
	end
	simulation.grid[x][y] = simulation.CONCRETE
end

function simulation.spawnplant(x,y) 
	if not simulation.inbounds(x,y) then
		return
	end
	simulation.grid[x][y] = simulation.PLANT
	simulation.additionalinfo[x][y].growthtimer = simulation.plantgrowtime
	simulation.additionalinfo[x][y].firehp = simulation.defaultfirehp
end


function simulation.spawnfire( x,y )
	if not simulation.inbounds(x,y) then
		return
	end
	simulation.grid[x][y] = simulation.FIRE
	simulation.additionalinfo[x][y].fuel = simulation.defaultfuel
end


function simulation.erase(x,y)
	if not simulation.inbounds(x,y) then
		return
	end
	simulation.grid[x][y] = simulation.NOTHING
end

function simulation.count_type(type)
	local count = 0
	for x = 1,simulation.size.x do
			for y=1,simulation.size.y do
				if simulation.grid[x][y] == type then
					count = count + 1
				end
		end
	end
	return count
end